from mpm.model import *
import shutil
import pathlib
import logging
import os
import semantic_version

logger = logging.getLogger('mpm.server')

class Server:
    def __init__(self, name, config):
        self.name = name
        self.config = config
        self.path = config['path']
        self.pluginPath = self.path+'/plugins'

    def exists(self):
        return os.path.exists(self.path) and os.path.exists(self.pluginPath)

    def plugins(self):
        return [PluginSpec(p['name'], p.get('version', '*')) for p in self.config['plugins']]

    def add_plugin(self, pluginSpec):
        for plugin in self.config['plugins']:
            if plugin['name'] == pluginSpec.name:
                raise KeyError("Cannot add plugin multiple times.")
        self.config['plugins'].append({'name': pluginSpec.name, 'version': str(pluginSpec.versionSpec)})

    def pluginStates(self, repos):
        managedPluginFilenames = []
        for plugin in self.plugins():
            logger.debug("Checking %r", plugin)
            compatibleVersions = []

            currentVersion = self.currentVersionForPlugin(plugin.name)
            logger.debug('Current version: %s', currentVersion)
            if currentVersion is not None:
                if plugin.versionSpec.match(currentVersion.version):
                    compatibleVersions.append(currentVersion)
                managedPluginFilenames.append(os.path.basename(currentVersion.path))

            for repo in repos:
                for repoPlugin in repo.plugins():
                    if repoPlugin.name == plugin.name:
                        logger.debug("%r -> %r", repoPlugin.version, plugin.versionSpec)
                        logger.debug(plugin.versionSpec.match(repoPlugin.version))
                        if plugin.versionSpec.match(repoPlugin.version):
                            compatibleVersions.append(repoPlugin)
            if len(compatibleVersions) == 0:
                logger.debug('No compatible versions found for %s', plugin)
                yield MissingVersions(plugin)
            else:
                preferredVersion = list(reversed(sorted(compatibleVersions)))[0]
                logger.debug('Wanted %r, found %r', plugin, preferredVersion)
                if currentVersion is not None and preferredVersion == currentVersion:
                    logger.debug("Already installed: %r", currentVersion)
                    yield Installed(plugin, currentVersion)
                else:
                    logger.debug("Update available: %r -> %r", plugin, preferredVersion)
                    yield Outdated(preferredVersion, currentVersion, plugin.versionSpec)

        otherPlugins = os.listdir(self.pluginPath)
        for pluginFile in otherPlugins:
            if os.path.isfile(os.path.join(self.pluginPath, pluginFile)) and pluginFile not in managedPluginFilenames:
                logger.debug("Unmanaged file: %s", pluginFile)
                yield UnmanagedFile(pluginFile)

    def currentVersionForPlugin(self, pluginName):
        plugins = os.listdir(self.pluginPath)
        for pluginJar in plugins:
            if pluginJar.startswith(pluginName) and pluginJar.endswith(".jar"):
                try:
                    return Plugin(os.path.join(self.pluginPath, pluginJar))
                except ValueError:
                    continue

    def syncToVersion(self, plugin):
        dest = os.path.join(self.pluginPath, plugin.normalizedFilename())
        logger.debug("Syncing %s to %s", dest, plugin)
        current = self.currentVersionForPlugin(plugin.name)
        if current is not None:
            logger.debug("Removing current version %r", current)
            current.unlink()

        try:
            os.link(plugin.path, dest)
            logger.debug("Plugin synced")
        except Exception:
            logger.debug("Hardlink not possible. Copying instead.")
            shutil.copyfile(plugin.path, dest)
